# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2022, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

import os
import pytest
from fluids.numerics import assert_close, assert_close1d
try:
    import rdkit
    from rdkit import Chem
except:
    rdkit = None
from thermo import Chemical
from thermo.group_contribution import Wilson_Jasperson

@pytest.mark.rdkit
@pytest.mark.skipif(rdkit is None, reason="requires rdkit")
def test_Wilson_Jasperson():
    # Two points in Poling et al.
    Tc, Pc, _, _ = Wilson_Jasperson('CCC1=CC=CC=C1O', Tb=477.67, second_order=True)
    assert_close(Tc, 693.5671723593391)
    assert_close(Pc, 3743819.6667063655)    
    
    Tc, Pc, _, _ = Wilson_Jasperson('CCC1=CC=CC=C1O', Tb=477.67, second_order=False)
    
    assert_close(Tc, 702.8831365703206)
    assert_close(Pc, 3794106.4902720796)
    
    # Had a bug identifying an amine group here
    c = Chemical('aniline')
    Tc, Pc, _, _ = Wilson_Jasperson(c.rdkitmol, Tb=457.4)
    assert_close(Tc, 705.7480487320958)
    assert_close(Pc, 5247025.774965471)
    
    c = Chemical('tetramethylthiuram disulfide')
    Tc, Pc, _, _ = Wilson_Jasperson(c.rdkitmol, Tb=580.6)
    assert_close(Tc, 807.2810024378236)
    assert_close(Pc, 2555268.114365961)
    
    # Osmium - can't work as a pure component
    c = Chemical('7440-04-2')
    Tc, Pc, missing_Tc_increments, missing_Pc_increments = Wilson_Jasperson(c.rdkitmol, Tb=5281.15)
    assert missing_Tc_increments
    assert missing_Pc_increments


    # Can't make it match no matter what I do, but nothing looks like an issue
    # c = Chemical('acetic acid')
    # Wilson_Jasperson_first_order(c.rdkitmol, Tb=391.2),584.6 
    
    # c = Chemical('2-nonanone')
    # Wilson_Jasperson_first_order(c.rdkitmol, Tb=467.7),651.8

@pytest.mark.rdkit
@pytest.mark.skipif(rdkit is None, reason="requires rdkit")
def test_Wilson_Jasperson_paper():

    chemical_names = ["ethane", "cyclopropane", "propane", "2-methylpropane", "butane", "cyclopentane", "2,2-dimethylpropane", "pentane", "2-methylbutane", "cyclohexane", "methylcyclopentane", "2,3-dimethylbutane", "3-methylpentane", "2,2-dimethylbutane", "2-methylpentane", "hexane", "ethylcyclopentane", "methylcyclohexane", "cycloheptane", "3,3-dimethylpentane", "2,4-dimethylpentane", "2,2-dimethylpentane", "heptane", "2-methylhexane", "3-ethylpentane", "2,3-dimethylpentane", "3-methylhexane", "2,2,3-trimethylbutane", "cyclooctane", "1,trans-4-dimethylcyclohexane", "2-methyl-3-ethylpentane", "3-ethylhexane", "3-ethyl-3-methylpentane", "2,3,3-trimethylpentane", "3,3-dimethylhexane", "2,2,3-trimethylpentane", "2,3,4-trimethylpentane", "octane", "4-methylheptane", "2,2,4-trimethylpentane", "2-methylheptane", "2,5-dimethylhexane", "2,2-dimethylhexane", "3-methylheptane", "3,4-dimethylhexane", "2,3-dimethylhexane", "2,4-dimethylhexane", "1,3,5-Trimethylcyclohexane", "2,2,3,3-tetramethylpentane", "2,3,3,4-tetramethylpentane", "2-methyloctane", "2,2,5-trimethylhexane", "2,2,3,4-tetramethylpentane", "2,2-dimethylheptane", "2,2,4,4-tetramethylpentane", "nonane", "cis-decalin", "trans-decalin", "2,2,3,3-tetramethylhexane", "decane", "3,3,5-trimethylheptane", "2,2,5,5-tetramethylhexane", "undecane", "dodecane", "tridecane", "tetradecane", "pentadecane", "hexadecane", "heptadecane", "octadecane", "ethene", "1,2-propadiene", "propene", "1,3-butadiene", "trans-2-butene", "cis-2-butene", "2-methylpropene", "1-butene", "cyclopentene", "cis-2-pentene", "2-methyl-2-butene", "1-pentene", "benzene", "cyclohexene", "1-hexene", "toluene", "1-heptene", "styrene", "1,2-dimethylbenzene", "1,3-dimethylbenzene", "ethylbenzene", "1,4-dimethylbenzene", "1-octene", "indan", "1-ethyl-4-methylbenzene", "propylbenzene", "1,2,3-trimethylbenzene", "1,3,5-trimethylbenzene", "1,2,4-trimethylbenzene", "isopropylbenzene", "1-nonene", "naphthalene", "tetralin", "1,2,4,5-tetramethylbenzene", "p-cymene", "isobutylbenzene", "butylbenzene", "1,4-diethylbenzene", "4,7,7-trimethyl-3-norcarene", "(1S)-(-)-.alpha.-pinene", "(R)-(+)-limonene", "1-decene", "2-methylnaphthalene", "1-methylnaphthalene", "1,1'-biphenyl", "2,7-dimethylnaphthalene", "hexamethylbenzene", "1-dodecene", "1,1':2',1''-terphenyl", "ethyne", "1-propyne", "bromotrifluoromethane", "chlorotrifluoromethane", "dichlorodifluoromethane", "trichlorofluoromethane", "carbon tetrachloride", "bromodifluoromethane", "chlorodifluoromethane", "dichlorofluoromethane", "chloroform", "trifluoromethane", "dichloromethane", "difluoromethane", "chloromethane", "fluoromethane", "chlorotrifluoroethene", "chloropentafluoroethane", "1,1,2-trichlorotrifluoroethane", "tetrafluoroethene", "hexafluoroethane", "1-chloro-2,2-difluoroethene", "1-chloro-1,2,2,2-tetrafluoroethane", "1,1-dichloro-2,2,2-trifluoroethane", "1,2-dichloro-1,1,2-trifluoroethane", "pentafluoroethane", "1-chloro-2,2,2-trifluoroethane", "1,1,2,2-tetrafluoroethane", "1,1,1,2-tetrafluoroethane", "1-chloro-1,1-difluoroethane", "1,1-dichloro-1-fluoroethane", "1,1,1-trifluoroethane", "1,2-dichloroethane", "1,1-dichloroethane", "1,1-difluoroethane", "bromoethane", "chloroethane", "fluoroethane", "1,2-dichlorohexafluoropropane", "octafluoropropane", "2H-perfluoropropane", "1,1,1,2,3,3-hexafluoropropane", "1,1,1,2,2-pentafluoropropane", "1,3-dichloropropane", "1,2-dichloropropane", "1-bromopropane", "1-chloropropane", "2-chloropropane", "octafluorocyclobutane", "perfluorobutane", "1,1,1,2,2,3,3,4-octafluorobutane", "2-chlorobutane", "1-chlorobutane", "perfluoropentane", "1-chloropentane", "tert-pentyl chloride", "perfluorobenzene", "perfluorohexane", "pentafluorobenzene", "1,2,4,5-tetrafluorobenzene", "1,3-dichlorobenzene", "1,3-difluorobenzene", "1,4-difluorobenzene", "1,2-difluorobenzene", "chlorobenzene", "fluorobenzene", "1-chlorohexane", "octafluorotoluene", "perfluoromethylcyclohexane", "perfluoroheptane", "4-fluorotoluene", "octadecafluorooctane", "1-chlorooctane", "methanol", "ethanol", "2-propenol", "2-propanol", "1-propanol", "1,2-propanediol", "2-methoxyethanol", "1-butanol", "2-methyl-1-propanol", "2-methyl-2-propanol", "2-butanol", "1,4-butanediol", "cyclopentanol", "1-pentanol", "2-methyl-2-butanol", "2-pentanol", "3-methyl-1-butanol", "2-methyl-1-butanol", "3-pentanol", "3-methyl-2-butanol", "2-propoxyethanol", "2-(2-methoxyethoxy)ethanol", "phenol", "cyclohexanol", "3-hexanol", "4-methyl-2-pentanol", "1-hexanol", "2-methyl-2-pentanol", "3-methyl-3-pentanol", "2-hexanol", "4-methyl-1-pentanol", "2-methyl-1-pentanol", "2-methyl-3-pentanol", "2-butoxyethanol", "3,6-dioxa-1-octanol", "2-methylphenol", "3-methylphenol", "4-methylphenol", "1-heptanol", "3-heptanol", "2-heptanol", "4-heptanol", "1-butoxy-2-propanol", "propyl carbitol", "2-ethylphenol", "3-ethylphenol", "1-phenylethanol", "2,6-xylenol", "4-ethylphenol", "2,4-dimethylphenol", "2,3-xylenol", "3,5-dimethylphenol", "3,4-xylenol", "2,5-xylenol", "2-octanol", "4-methyl-3-heptanol", "2-ethyl-1-hexanol", "1-octanol", "5-methyl-3-heptanol", "2-(2-butoxyethoxy)ethanol", "1-nonanol", "4-nonanol", "2-nonanol", "1-decanol", "2-decanol", "1-undecanol", "1-dodecanol", "oxirane", "methoxymethane", "epoxypropane", "methoxyethane", "dimethoxymethane", "furan", "ethyl vinyl ether", "tetrahydrofuran", "1,4-dioxane", "1-methoxypropane", "diethyl ether", "2-methoxypropane", "1,2-dimethoxyethane", "2-methylfuran", "3,4-dihydro-2H-pyran", "2-methyltetrahydrofuran", "tetrahydropyran", "methyl tert-butyl ether", "1-ethoxypropane", "butyl methyl ether", "butyl vinyl ether", "tert-amyl methyl ether", "tert-butyl ethyl ether", "isopropyl ether", "dipropyl ether", "methyl pentyl ether", "1,1-diethoxyethane", "methoxybenzene", "dibutyl ether", "dibenzofuran", "diphenyl ether", "propanone", "butanone", "cyclopentanone", "3-methyl-2-butanone", "2-pentanone", "3-pentanone", "cyclohexanone", "mesityl oxide", "2-hexanone", "3-hexanone", "4-methyl-2-pentanone", "4-heptanone", "3-heptanone", "2-heptanone", "acetophenone", "3-octanone", "4-octanone", "2-octanone", "2-nonanone", "5-nonanone", "4-nonanone", "3-nonanone", "5-decanone", "4-decanone", "2-decanone", "3-decanone", "6-undecanone", "2-undecanone", "3-undecanone", "2-dodecanone", "benzophenone", "ethanal", "propanal", "butanal", "pentanal", "hexanal", "heptanal", "octanal", "decanal", "acetic acid", "propanoic acid", "2-methylpropanoic acid", "butanoic acid", "pentanoic acid", "3-methylbutanoic acid", "hexanoic acid", "heptanoic acid", "octanoic acid", "2-ethylhexanoic acid", "nonanoic acid", "decanoic acid", "methyl methanoate", "methyl ethanoate", "ethyl methanoate", "dimethyl carbonate", "vinyl acetate", "gamma-butyrolactone", "methyl propanoate", "ethyl ethanoate", "propyl methanoate", "ethyl propanoate", "isopropyl ethanoate", "methyl butanoate", "propyl ethanoate", "methyl isobutyrate", "isobutyl ethanoate", "ethyl butanoate", "butyl ethanoate", "propyl propanoate", "cellosolve acetate", "1-methoxy-2-propyl acetate", "propyl butanoate", "propyl isobutyrate", "pentyl ethanoate", "ethyl isovalerate", "isobutyl propanoate", "ethyl pentanoate", "isoamyl acetate", "2-butoxyethyl acetate", "ethyl octanoate", "ethyl nonanoate", "methylamine", "ethylamine", "dimethylamine", "1,2-ethanediamine", "N,N-dimethylformamide", "propylamine", "trimethylamine", "isopropylamine", "pyrrole", "pyrrolidine", "2-butanamine", "diethylamine", "tert-butylamine", "butylamine", "piperidine", "aniline", "diisopropylamine", "dipropylamine", "triethylamine", "N,N-dimethylaniline", "diisobutylamine", "dibutylamine", "N,N,2-trimethylaniline", "pyridine", "3-methylpyridine", "2-methylpyridine", "4-methylpyridine", "2,6-dimethylpyridine", "3,5-dimethylpyridine", "2,3-dimethylpyridine", "3,4-dimethylpyridine", "2,4-dimethylpyridine", "2,5-dimethylpyridine", "acetonitrile", "propenenitrile", "propanenitrile", "butanenitrile", "pentanenitrile", "hexanenitrile", "benzonitrile", "octanenitrile", "methanethiol", "2-thiapropane", "ethanethiol", "1-propanethiol", "thiophene", "ethyl thiolacetate", "tetrahydrothiophene", "3-thiapentane", "1-butanethiol", "thianaphthene", "dibenzothiophene", "pentafluorodimethyl ether", "2,2,2-trifluoroethanol", "trifluoromethoxymethane", "hexafluorooxetane", "perfluoro(dimethoxymethane)", "1,1,2,2-tetrafluoroethyl trifluoromethyl ether", "1-chloro-2,2,2-trifluoro difluoromethyl ether", "1,2,2,2-tetrafluoroethyl difluoromethyl ether", "methyl pentafluoroethyl ether", "1,2,2,3-tetrafluoro-1-propanol", "methyl 2,2,2-trifluoroethyl ether", "perfluorotetrahydrofuran", "heptafluoropropyl trifluoromethyl ether", "4,4,5,5-tetrafluoro-2-(trifluoromethyl)- 1,3-dioxolane", "heptafluoro-1,4-dioxane", "1,1-bis(difluoromethoxy)tetrafluoroethane", "3,3,4,4,4-pentafluoro-2-butanone", "methyl perfluoroisopropyl ether", "heptafluoropropyl methyl ether", "ethyl pentafluoroethyl ether", "perfluorovaleric acid", "methyl perfluoroisopropyl ketone", "tert-perfluorobutyl methyl ether", "1,5,5-trihydrooctafluoropentanol", "1,7,7-trihydrododecafluoroheptanol"]
    NBPs = [184.6, 240.3, 231, 261.4, 272.6, 322.4, 282.6, 309.2, 301.1, 353.9, 345, 331.2, 336.4, 322.9, 333.4, 341.9, 376.6, 374.1, 392, 359.2, 353.7, 352.4, 371.6, 363.2, 366.6, 363, 365, 354.1, 424, 392.5, 388.8, 391.7, 391.5, 387.8, 385.1, 383, 386.6, 398.8, 390.8, 372.4, 390.8, 382.3, 380, 392.1, 390.9, 388.8, 382.6, 413.7, 413.4, 414.6, 416.4, 397.1, 406.2, 405.9, 395.4, 423.9, 468.8, 460.2, 432.5, 447.3, 428.8, 410.2, 468.5, 489.5, 507.5, 526.4, 544, 560, 576, 590, 169.3, 238.7, 225.5, 268.7, 274.1, 276.8, 266.6, 266.8, 317.4, 310, 311.6, 303.3, 353.3, 356.1, 336.7, 383.8, 366.6, 418.6, 417.6, 412.3, 409.4, 411.5, 394.5, 450.7, 435.1, 432.5, 449.2, 437.8, 442.5, 425.5, 419.6, 491.15, 480.5, 469.6, 450.2, 445.9, 456.4, 456.8, 445, 430, 450.1, 444.3, 514.2, 517.6, 528.3, 538, 536.9, 486.5, 609, 189.5, 249.9, 215.3, 191.8, 243.4, 296.8, 349.9, 257.8, 232.3, 282, 334.4, 191, 313, 221.5, 249.3, 194.7, 244.6, 234, 320, 198, 195.1, 254.5, 261.2, 301, 302, 225.2, 279.3, 250.2, 247.1, 263.6, 305.1, 226.3, 356.6, 330.5, 249.1, 311.6, 286.1, 235.7, 310, 236.2, 256, 279.7, 254.8, 393.5, 369.4, 343.7, 319.7, 308, 267.2, 271.2, 300.5, 341, 351.6, 302.1, 381, 358.2, 353.4, 330.3, 358.9, 363.4, 446.2, 356.2, 362, 367.1, 404.9, 357.9, 408.2, 377.8, 349.5, 355.1, 389.8, 379, 456.4, 337.8, 351.5, 370.2, 355.5, 370.3, 460.3, 397.5, 390.8, 380.9, 355.6, 372.6, 502.9, 414, 411, 375.1, 392.3, 404.5, 401.8, 388.6, 385.2, 424.5, 466, 455, 434, 407.6, 404, 430, 395, 395.1, 412.4, 425, 421, 399.7, 444.3, 468, 464.2, 475.4, 475.1, 449, 429.5, 432.5, 428, 443.3, 488, 479.5, 490.6, 475.7, 474.2, 492, 483.2, 491.1, 493.4, 499.5, 483.6, 453, 428.2, 456, 468.1, 427, 504.5, 486.4, 465.6, 471, 504.2, 482, 517, 538, 283.7, 248.6, 307.5, 280.4, 315.4, 304.7, 308.8, 339, 374.5, 311.8, 307.8, 304, 358, 337, 358.8, 353, 361, 328.2, 336.3, 343.3, 367.1, 359.5, 346, 341.4, 363.3, 372.4, 376.2, 426.9, 415, 559, 531.5, 329.3, 352.7, 403.6, 367.1, 375.2, 375, 428.6, 402.8, 400.3, 396.5, 389.1, 416.8, 420.4, 423.8, 475.2, 438.6, 438.7, 446, 467.7, 461, 460.7, 461, 476.7, 478.7, 485.5, 478, 499, 504, 500, 519.7, 579, 294, 321.5, 347.8, 375.7, 401.6, 427, 442, 480, 391.2, 414.2, 427.3, 436.1, 459.2, 449.2, 477.5, 496, 510.7, 500.7, 527.2, 541.2, 304.9, 330.1, 327.3, 363.5, 345.7, 478, 352.7, 350.3, 354.1, 372.2, 361.8, 375.7, 374.7, 365.4, 390.2, 393.8, 399.2, 396, 429.7, 420, 416.2, 407, 421.7, 408.5, 410.1, 417.6, 414.7, 466, 480, 500.2, 266.7, 290, 280.4, 390.7, 426, 322.2, 276.4, 305.5, 403.1, 359.5, 336.2, 328.5, 317.3, 350.7, 379.2, 457.4, 357, 382.5, 362.2, 466.8, 412, 432.8, 458.5, 388.5, 417, 402.5, 418.3, 417.2, 445, 434.3, 452.2, 431.4, 430, 354.8, 350.4, 370.5, 390.9, 414, 436, 464.1, 475.5, 279.1, 310.3, 308.2, 340.9, 357.3, 386.9, 394.5, 365.2, 371, 494, 604.6, 238.6, 347.1, 239.3, 244.6, 263.3, 272.4, 322.5, 296.5, 278.7, 386.1, 304.8, 272.7, 280, 305, 312.7, 320, 314.4, 302.6, 307.4, 301.3, 414, 328.8, 326.5, 414, 444.2, ]
    expect_Tcs = [304.3, 408.1, 368.9, 406.2, 423.6, 512, 428.7, 469.1, 456.8, 547.2, 533.4, 491.8, 499.5, 479.4, 495, 507.7, 568.8, 565, 592, 523.1, 515.1, 513.2, 541.1, 528.9, 533.9, 528.6, 531.5, 515.7, 627, 580.5, 556.3, 560.4, 560.1, 554.8, 551, 548, 553.1, 570.6, 559.1, 532.8, 559.1, 547, 543.7, 561, 559.3, 556.3, 547.4, 600.3, 581.9, 583.6, 586.2, 559, 571.8, 571.4, 556.6, 596.7, 690.6, 677.9, 599.7, 620.3, 594.6, 568.8, 640.6, 660.7, 676.6, 693.8, 709.2, 722.5, 735.9, 746.7, 283.1, 390.8, 364.5, 426.6, 430.4, 434.7, 418.6, 418.9, 509.9, 474.6, 477.1, 464.4, 564, 556.2, 504.1, 596.3, 537.9, 640.7, 633.4, 625.3, 620.9, 624.1, 568.3, 696.7, 645.9, 642, 666.9, 649.9, 656.9, 631.6, 594.4, 755.3, 725.5, 683.7, 655.5, 649.2, 664.5, 665.1, 660.7, 638.4, 650.6, 619.7, 772.9, 778, 784.1, 792.2, 755.6, 660, 856.8, 321.7, 409.1, 342.8, 306.2, 386.8, 469.4, 550.9, 415.8, 375.7, 453.8, 535.4, 310.5, 510.4, 365.3, 414.9, 326, 384, 356.1, 483.3, 312.2, 298, 404.2, 401.5, 460.9, 462.4, 347.6, 433.9, 390.4, 385.5, 414.1, 477.2, 357.1, 564.3, 523, 397.9, 500.3, 460.6, 381.4, 452.4, 346.9, 379.1, 417.9, 384.2, 606.5, 569.4, 536.5, 500.3, 482, 398.8, 385.4, 433.3, 520.5, 536.7, 417.4, 568.8, 534.8, 527.9, 445.2, 541, 553.1, 687.6, 553.5, 562.5, 570.4, 633.4, 562.4, 597.4, 544.6, 478.7, 483.6, 597.2, 489.9, 645.2, 513.4, 522.7, 544, 518.5, 540.1, 654.6, 570.4, 559.9, 545.7, 509.5, 533.8, 704, 623.2, 596.9, 544.8, 569.7, 587.5, 583.5, 564.4, 559.4, 606.7, 656.3, 687.8, 639.9, 581.7, 576.6, 613.7, 563.7, 563.9, 588.6, 606.5, 600.8, 570.4, 624.9, 649.3, 687, 703.6, 703.1, 630.6, 603.2, 607.4, 601.1, 614.2, 667.6, 696.2, 712.3, 690.7, 688.5, 714.4, 701.6, 713.1, 716.4, 725.3, 702.2, 626.8, 592.5, 631, 647.7, 590.8, 681.2, 663.8, 635.4, 642.8, 679.2, 649.3, 688.1, 707.9, 483.6, 398.1, 505.1, 436.8, 480.5, 497.1, 473.8, 539.7, 581.7, 474, 467.9, 462.1, 533.5, 533.4, 561.8, 547.1, 559.5, 488.2, 500.3, 510.7, 539.5, 524.4, 504.7, 498, 530, 543.2, 539.9, 648.8, 585, 845.5, 775, 516.4, 539.4, 629.6, 549, 561.1, 560.8, 652.4, 595, 586.7, 581.2, 570.3, 599.9, 605.1, 609.9, 710.5, 620.8, 620.9, 631.3, 651.8, 642.5, 642.1, 642.5, 654.9, 657.7, 667, 656.7, 676.5, 683.2, 677.8, 695.8, 833, 469, 499, 527.1, 557.3, 584.4, 610.5, 621.8, 656, 584.6, 606.7, 614.6, 627.3, 649.6, 635.4, 665.2, 681.1, 692, 678.5, 705.5, 715.8, 474.4, 507.6, 497, 547.4, 519.6, 750.2, 530.1, 526.5, 526.3, 548, 532.7, 553.1, 551.7, 538, 563.9, 569.1, 576.9, 572.3, 611.4, 597.6, 591.3, 578.2, 599.1, 580.4, 582.6, 593.3, 589.2, 643.1, 652.7, 671.6, 435.7, 459.6, 444.3, 599, 641.7, 497.3, 426.6, 471.5, 650, 566.6, 506.9, 495.3, 478.4, 528.8, 582.4, 705.7, 517.3, 554.3, 524.8, 689.2, 577.5, 606.7, 664.2, 610.6, 639, 616.8, 641, 625.1, 666.7, 650.7, 677.5, 646.3, 644.2, 551.8, 536.9, 562.5, 580.8, 603.2, 624.1, 689.4, 659.7, 468.2, 503, 499.6, 536.6, 590.7, 584.7, 635.5, 560.3, 569.2, 781.6, 922.8, 360.5, 502.4, 368.8, 372.7, 374.4, 396.4, 471.6, 435, 412.3, 545.5, 459, 399.7, 392.3, 443, 454.2, 447.9, 458.5, 425.7, 439.6, 437.3, 554.2, 464.4, 453.5, 568.4, 582.8, ]
    missed = 0
    wrong = 0
    
    for name, Tb, expect in zip(chemical_names, NBPs, expect_Tcs):
        try:
            c = Chemical(name)
        except:
            missed += 1
            continue
        Tc, Pc, miss_Tc, miss_Pc = Wilson_Jasperson(c.rdkitmol, Tb)
        assert not miss_Tc
        assert not miss_Pc



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2020, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

import pytest
import thermo
from thermo import *
from thermo.coolprop import *
from thermo.phases import IAPWS95Gas, IAPWS95Liquid
from thermo.chemical_package import iapws_correlations
from fluids.numerics import *
from chemicals.iapws import iapws95_Tc, iapws95_Pc
from thermo.test_utils import mark_plot_unsupported
from math import *
import json
import os
import numpy as np
try:
    import matplotlib.pyplot as plt
except:
    pass

fluid = 'water'
pure_surfaces_dir = os.path.join(thermo.thermo_dir, '..', 'surfaces', 'iapws95')


def test_iapws95_basic_flash():
    liquid = IAPWS95Liquid(T=300, P=1e5, zs=[1])
    gas = IAPWS95Gas(T=300, P=1e5, zs=[1])
    flasher = FlashPureVLS(iapws_constants, iapws_correlations, gas, [liquid], [])
    PT = flasher.flash(T=300, P=1e6)

    assert_close(PT.rho_mass(), 996.9600226949985, rtol=1e-10)
    assert isinstance(PT.liquid0, IAPWS95Liquid)

    TV = flasher.flash(T=300, V=PT.V())
    assert_close(TV.P, PT.P, rtol=1e-10)
    assert isinstance(TV.liquid0, IAPWS95Liquid)

    PV = flasher.flash(P=PT.P, V=PT.V())
    assert_close(PV.T, PT.T, rtol=1e-13)
    assert isinstance(PV.liquid0, IAPWS95Liquid)

    TVF = flasher.flash(T=400.0, VF=1)
    assert_close(TVF.P, 245769.3455657166, rtol=1e-13)

    PVF = flasher.flash(P=TVF.P, VF=1)
    assert_close(PVF.T, 400.0, rtol=1e-13)

    assert_close(flasher.flash(P=1e5, H=flasher.flash(T=273.15, P=1e5).H()).T, 273.15, rtol=1e-10)



    T1 = 350 + 273.15
    P1 = 100*1e5
    P2 = 1e5
    stage_1 = flasher.flash(P=P1, T=T1)
    stage_2 = flasher.flash(P=P2, S=stage_1.S())
    assert_close(stage_2.VF, 0.7666960540476834)
    
    
def test_iapws95_basic_flashes_no_hacks():
    liquid = IAPWS95Liquid(T=300, P=1e5, zs=[1])
    gas = IAPWS95Gas(T=300, P=1e5, zs=[1])
    flasher = FlashPureVLS(iapws_constants, iapws_correlations, gas, [liquid], [])
    flasher.VL_only_IAPWS95 = False

    PT = flasher.flash(T=300, P=1e6)

    assert_close(PT.rho_mass(), 996.9600226949985, rtol=1e-10)

    TV = flasher.flash(T=300, V=PT.V())
    assert_close(TV.P, PT.P, rtol=1e-10)

    PV = flasher.flash(P=PT.P, V=PT.V())
    assert_close(PV.T, PT.T, rtol=1e-13)

    PH = flasher.flash(H=PT.H(), P=1e6)
    assert_close(PH.T, 300)

    PS = flasher.flash(S=PT.S(), P=1e6)
    assert_close(PS.T, 300)

    assert_close(flasher.flash(P=1e5, H=flasher.flash(T=273.15, P=1e5).H()).T, 273.15, rtol=1e-10)
    
    # Case where the density solution was failing
    T = 250.49894958453197
    P = 10595.601792776019
    base = flasher.flash(T=T, P=P)
    new = flasher.flash(P=P, S=base.S())
    assert_close(new.T, base.T)


@pytest.mark.plot
@pytest.mark.slow
@pytest.mark.parametric
@pytest.mark.parametrize("variables", ['VPT', 'VTP',
                                       'PHT', 'PST', 'PUT',
                                       'VUT', 'VST', 'VHT',
                                          'TSV',
                                            'THP', 'TUP',
                                       ])
def test_plot_IAPWS95(variables):
    spec0, spec1, check_prop = variables
    plot_name = variables[0:2]
    eos = IAPWS95
    T, P = 298.15, 101325.0
    zs = [1.0]
    liquid = IAPWS95Liquid(T=300, P=1e5, zs=[1])
    gas = IAPWS95Gas(T=300, P=1e5, zs=[1])

    flasher = FlashPureVLS(constants=iapws_constants, correlations=iapws_correlations,
                       gas=gas, liquids=[], solids=[])
    flasher.TPV_HSGUA_xtol = 1e-13
    
    flash_spec = frozenset([spec0, spec1])
    inconsistent = flash_spec in (frozenset(['T', 'H']), frozenset(['T', 'U']),
                                  frozenset(['T', 'S']), # blip issues
                                  frozenset(['V', 'P']), # 4 degree water blip
                                  frozenset(['V', 'S']), frozenset(['V', 'H']),
                                  frozenset(['V', 'U']), # Fun
                                  )

    res = flasher.TPV_inputs(zs=[1.0], pts=200, spec0='T', spec1='P', 
                             check0=spec0, check1=spec1, prop0=check_prop,
                           trunc_err_low=1e-13,
                           trunc_err_high=1, color_map=cm_flash_tol(),
                           show=False, verbose=not inconsistent)

    matrix_spec_flashes, matrix_flashes, errs, plot_fig = res

    path = os.path.join(pure_surfaces_dir, fluid, plot_name)
    if not os.path.exists(path):
        os.makedirs(path)
        
    tol = 5e-12

    key = '%s - %s - %s' %(plot_name, eos.__name__, fluid)

    if inconsistent:
        spec_name = spec0 + spec1
        mark_plot_unsupported(plot_fig, reason='EOS is inconsistent for %s inputs' %(spec_name))
        tol = 1e300
    if flash_spec == frozenset(['T', 'V']):
        tol = 1e-5

    plot_fig.savefig(os.path.join(path, key + '.png'))
    plt.close()

    max_err = np.max(np.abs(errs))
    assert max_err < tol

# test_plot_IAPWS95('VUT')
# test_plot_IAPWS95('PUT')


@pytest.mark.plot
@pytest.mark.slow
@pytest.mark.parametric
@pytest.mark.parametrize("variables", ['PSS', 'TSS'])
def test_plot_IAPWS95_VF(variables):
    spec0, spec1, check_prop = variables
    plot_name = variables[0:2] + 'VF'
    eos = IAPWS95
    T, P = 298.15, 101325.0
    zs = [1.0]
    liquid = IAPWS95Liquid(T=300, P=1e5, zs=[1])
    gas = IAPWS95Gas(T=300, P=1e5, zs=[1])

    flasher = FlashPureVLS(constants=iapws_constants, correlations=iapws_correlations,
                       gas=gas, liquids=[liquid], solids=[])
    flasher.TPV_HSGUA_xtol = 1e-13
    
    flash_spec = frozenset([spec0, spec1])
    res = flasher.TPV_inputs(zs=[1.0], pts=200, spec0='T', spec1='VF',
                             check0=spec0, check1=spec1, prop0=check_prop,
                             Tmax=iapws95_Tc*(1-1e-12),
                             Pmax=iapws95_Pc*(1-1e-12),
                           trunc_err_low=1e-13,
                           trunc_err_high=1, color_map=cm_flash_tol(),
                           show=False, verbose=True)

    matrix_spec_flashes, matrix_flashes, errs, plot_fig = res

    path = os.path.join(pure_surfaces_dir, fluid, plot_name)
    if not os.path.exists(path):
        os.makedirs(path)
        
    tol = 5e-10

    key = '%s - %s - %s' %(plot_name, eos.__name__, fluid)

    plot_fig.savefig(os.path.join(path, key + '.png'))
    plt.close()

    max_err = np.max(np.abs(errs))
    assert max_err < tol



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2016, Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

import pytest
import numpy as np
import json
import pandas as pd
from thermo.interface import *
from chemicals.utils import property_mass_to_molar, zs_to_ws
from chemicals.dippr import EQ106
from thermo.interface import VDI_TABULAR
from chemicals.identifiers import check_CAS
from fluids.numerics import assert_close, assert_close1d, linspace, derivative
from thermo.volume import VolumeLiquid, VDI_PPDS
from thermo.utils import POLY_FIT
from thermo.interface import SurfaceTensionMixture, DIGUILIOTEJA, LINEAR, WINTERFELDSCRIVENDAVIS

@pytest.mark.meta_T_dept
def test_SurfaceTension():
    # Ethanol, test as many methods as possible at once
    EtOH = SurfaceTension(Tb=351.39, Tc=514.0, Pc=6137000.0, Vc=0.000168, Zc=0.24125, omega=0.635, StielPolar=-0.01266, CASRN='64-17-5')
    methods = list(EtOH.all_methods)
    methods_nontabular = list(methods)
    methods_nontabular.remove(VDI_TABULAR)
    sigma_calcs = []
    for i in methods_nontabular:
        EtOH.method = i
        sigma_calcs.append(EtOH.T_dependent_property(305.))

    sigma_exp = [0.021222422444285592, 0.02171156653650729, 0.02171156653650729, 0.021462066798796135, 0.02140008, 0.038055725907414066, 0.03739257387107131, 0.02645171690486362, 0.03905907338532845, 0.03670733205970745]

    assert_close1d(sorted(sigma_calcs), sorted(sigma_exp), rtol=1e-6)
    assert_close(EtOH.calculate(305., VDI_TABULAR), 0.021533867879206747, rtol=1E-4)

    # Test that methods return None
    EtOH.extrapolation = None
    for i in methods:
        EtOH.method = i
        assert EtOH.T_dependent_property(5000) is None


    assert SurfaceTension.from_json(EtOH.as_json()) == EtOH


    EtOH.method = 'VDI_TABULAR'
    EtOH.tabular_extrapolation_permitted = False
    assert None == EtOH.T_dependent_property(700.)

    with pytest.raises(Exception):
        EtOH.test_method_validity(300, 'BADMETHOD')

    # Test Aleem

    CH4 = SurfaceTension(Tb=111.65, Cpl=property_mass_to_molar(2465.,16.04246), Hvap_Tb=510870., MW=16.04246, Vml=3.497e-05)
    assert_close(CH4.T_dependent_property(90), 0.016704545538936296)

    assert not CH4.test_method_validity(600, 'Aleem')
    assert CH4.test_method_validity(100, 'Aleem')

@pytest.mark.meta_T_dept
def test_SurfaceTension_water_iapws():
    water = SurfaceTension(CASRN="7732-18-5", MW=18.01528, Tb=373.124, Tc=647.14, Pc=22048320.0, Vc=5.6e-05, Zc=0.229, omega=0.344, StielPolar=0.0232, Hvap_Tb=2256470., extrapolation="DIPPR106_AB", method="IAPWS")
    assert water.method == 'IAPWS'
    assert_close(water(400), 0.05357792640201927)
    
    assert water(3) is not None
    assert water(3) > 0



@pytest.mark.meta_T_dept
def test_SurfaceTensionJasperMissingLimits():
    obj = SurfaceTension(CASRN='110-01-0')
    assert_close(obj.calculate(obj.JASPER_Tmax, 'JASPER'), 0, atol=1e-10)

    obj = SurfaceTension(CASRN='14901-07-6')
    assert_close(obj.calculate(obj.JASPER_Tmax, 'JASPER'), 0, atol=1e-10)

@pytest.mark.meta_T_dept
def test_SurfaceTensionVDITabularMissingZeroLimits():
    obj = SurfaceTension(CASRN='7782-41-4')
    assert_close(obj.calculate(144.41, 'VDI_TABULAR'), 0, atol=1e-10)

@pytest.mark.fitting
@pytest.mark.meta_T_dept
def test_SurfaceTension_fitting0_yaws():
    A, Tc, n = 117.684, 7326.47, 1.2222
    Ts = linspace(378.4, 7326.47, 10)
    props_calc = [EQ106(T, Tc, A, n, 0.0, 0.0, 0.0, 0) for T in Ts]
    res, stats = SurfaceTension.fit_data_to_model(Ts=Ts, data=props_calc, model='YawsSigma',
                          do_statistics=True, use_numba=False, model_kwargs={'Tc': Tc},
                          fit_method='lm')        
    assert stats['MAE'] < 1e-5


@pytest.mark.meta_T_dept
def test_SurfaceTension_no_polyfit():
    sigma_kwargs_no_polyfit = {'CASRN': '7727-37-9', 'MW': 28.0134, 'Tb': 77.355, 'Tc': 126.2, 'Pc': 3394387.5, 'Vc': 8.95e-05, 'Zc': 0.2895282296391198, 'omega': 0.04, 'StielPolar': 0.009288558616105336, 'Hvap_Tb': 199160.50751339202, 'poly_fit': ()}
    created = SurfaceTension(**sigma_kwargs_no_polyfit)
    assert created(100) is not None

@pytest.mark.meta_T_dept
def test_SurfaceTension_exp_poly_fit_ln_tau():
    coeffs = [-1.2616237655927602e-05, -0.0004061873638525952, -0.005563382112542401, -0.04240531802937599, -0.19805733513004808, -0.5905741856310869, -1.1388001144550794, -0.1477584393673108, -2.401287527958821] 
    Tc = 647.096
    Tmin, Tmax = 233.22, 646.15

    # Create an object with no CAS, check a value
    good_obj = SurfaceTension(Tc=Tc, exp_poly_fit_ln_tau=(Tmin, Tmax, Tc, coeffs))
    expect = 0.07175344047522199
    assert_close(good_obj(300), expect, rtol=1e-12)
    
    # Create an object with a CAS, check a value
    good_obj2 = SurfaceTension(Tc=Tc, CASRN='7732-18-5', exp_poly_fit_ln_tau=(Tmin, Tmax, Tc, coeffs))
    assert_close(good_obj2(300), expect, rtol=1e-12)

    expect_der = -0.000154224581713238
    expect_der2 = -5.959538970287795e-07
    
    assert_close(good_obj.T_dependent_property_derivative(300), expect_der, rtol=1e-13)
    assert_close(good_obj.T_dependent_property_derivative(300, order=2), expect_der2, rtol=1e-13)


    assert SurfaceTension.from_json(good_obj2.as_json()) == good_obj2
    assert SurfaceTension.from_json(good_obj.as_json()) == good_obj
    assert eval(str(good_obj2)) == good_obj2
    assert eval(str(good_obj)) == good_obj
    
@pytest.mark.meta_T_dept
def test_SurfaceTension_exp_poly_ln_tau_extrapolate():
    coeffs = [1.1624065398371628, -1.9976745939643825]
    
    Tc = 647.096
    Tmin, Tmax = 233.22, 646.15
    good_obj = SurfaceTension(Tc=Tc, exp_poly_fit_ln_tau=(Tmin, Tmax, Tc, coeffs), extrapolation='EXP_POLY_LN_TAU2')
    assert_close(good_obj.calculate(1, good_obj.method), good_obj(1), rtol=1e-13)
    assert_close(good_obj.calculate(647, good_obj.method), good_obj(647), rtol=1e-13)

    assert_close(derivative(good_obj, 50, dx=50*1e-7), good_obj.T_dependent_property_derivative(50))
    
    assert_close(derivative(good_obj.T_dependent_property_derivative, 190, dx=190*1e-6), 
                 good_obj.T_dependent_property_derivative(190, order=2), rtol=1e-7)
    
    assert_close(good_obj.T_dependent_property_derivative(50), -0.0002405128589491164, rtol=1e-11)
    assert_close(good_obj.T_dependent_property_derivative(50, 2),6.541805875147306e-08, rtol=1e-11)
    
    # Floating-point errors pile up in this one
    coeffs = [-0.02235848200899392,  1.0064575672832703,  -2.0629066032890777 ]
    good_obj = SurfaceTension(Tc=Tc, exp_poly_fit_ln_tau=(Tmin, Tmax, Tc, coeffs), extrapolation='EXP_POLY_LN_TAU3')
    assert_close(good_obj.calculate(200, good_obj.method), good_obj(200), rtol=1e-7)
    assert_close(good_obj.calculate(647, good_obj.method), good_obj(647), rtol=1e-7)

    assert_close(derivative(good_obj, 50, dx=50*1e-7), good_obj.T_dependent_property_derivative(50))
    
    assert_close(derivative(good_obj.T_dependent_property_derivative, 190, dx=190*1e-6), 
                 good_obj.T_dependent_property_derivative(190, order=2), rtol=1e-7)
    
    
    assert_close(good_obj.T_dependent_property_derivative(50), -0.00019823412229632575, rtol=1e-11)
    assert_close(good_obj.T_dependent_property_derivative(50, 2), -1.136037983710264e-08, rtol=1e-11)
    
@pytest.mark.meta_T_dept
def test_SurfaceTension_exp_cheb_fit_ln_tau():
    coeffs = [-5.922664830406188, -3.6003367212635444, -0.0989717205896406, 0.05343895281736921, -0.02476759166597864, 0.010447569392539213, -0.004240542036664352, 0.0017273355647560718, -0.0007199858491173661, 0.00030714447101984343, -0.00013315510546685339, 5.832551964424226e-05, -2.5742454514671165e-05, 1.143577875153956e-05, -5.110008470393668e-06, 2.295229193177706e-06, -1.0355920205401548e-06, 4.690917226601865e-07, -2.1322112805921556e-07, 9.721709759435981e-08, -4.4448656630335925e-08, 2.0373327115630335e-08, -9.359475430792408e-09, 4.308620855930645e-09, -1.9872392620357004e-09, 9.181429297400179e-10, -4.2489342599871804e-10, 1.969051449668413e-10, -9.139573819982871e-11, 4.2452263926406886e-11, -1.9768853221080462e-11, 9.190537220149508e-12, -4.2949394041258415e-12, 1.9981863386142606e-12, -9.396025624219817e-13, 4.335282133283158e-13, -2.0410756418343112e-13, 1.0455525334407412e-13, -4.748978987834107e-14, 2.7630675525358583e-14]
    Tmin, Tmax, Tc = 233.22, 646.15, 647.096
    
    obj2 = SurfaceTension(Tc=Tc, exp_cheb_fit_ln_tau=((Tmin, Tmax, Tc, coeffs)))
    
    T = 500
    assert_close(obj2(T), 0.031264474019763455, rtol=1e-13)
    assert_close(obj2.T_dependent_property_derivative(T), -0.00023379922039411865, rtol=1e-13)
    assert_close(obj2.T_dependent_property_derivative(T, order=2), -1.0453010755999069e-07, rtol=1e-13)
    
@pytest.mark.meta_T_dept
def test_SurfaceTension_exp_stablepoly_fit_ln_tau():
    coeffs = [0.011399360373616219, -0.014916568994522095, -0.06881296308711171, 0.0900153056718409, 0.19066633691545576, -0.24937350547406822, -0.3148389292182401, 0.41171834646956995, 0.3440581845934503, -0.44989947455906076, -0.2590532901358529, 0.33869134876113094, 0.1391329435696207, -0.18195230788023764, -0.050437145563137165, 0.06583166394466389, 0.01685157036382634, -0.022266583863000733, 0.003539388708205138, -0.005171064606571463, 0.012264455189935575, -0.018085676249990357, 0.026950795197264732, -0.04077120220662778, 0.05786417011592615, -0.07222889554773304, 0.07433570330647113, -0.05829288696590232, -3.7182636506596722, -5.844828481765601]
    Tmin, Tmax, Tc = 233.22, 646.15, 647.096
    
    obj2 = SurfaceTension(Tc=Tc, exp_stablepoly_fit_ln_tau=(Tmin, Tmax, Tc, coeffs))
    
    T = 500
    assert_close(obj2(T), 0.03126447402046822, rtol=1e-13)
    assert_close(obj2.T_dependent_property_derivative(T), -0.0002337992205182661, rtol=1e-13)
    assert_close(obj2.T_dependent_property_derivative(T, order=2), -1.0453011134030858e-07, rtol=1e-13)

@pytest.mark.meta_T_dept
def test_SurfaceTension_EQ106_ABC_extrapolate():
    Tmin, Tc, A, B = 194.0, 592.5, 0.056, 1.32
    C = -0.01
    Tmax = 590.0
    
    good_obj = SurfaceTension(Tc=Tc, DIPPR106_parameters={'Test': {'Tmin': Tmin,
                                'Tmax': Tmax, 'Tc': Tc, 'A': A, 'B': B, 'C': C}},
                              extrapolation='DIPPR106_ABC')
    assert_close(good_obj(591), good_obj.calculate(591, good_obj.method), rtol=1e-13)
    assert_close(good_obj(50), good_obj.calculate(50, good_obj.method), rtol=1e-13)

    # Three derivatives
    assert_close(derivative(good_obj, 50, dx=50*1e-7), good_obj.T_dependent_property_derivative(50))
    
    assert_close(derivative(good_obj.T_dependent_property_derivative, 190, dx=190*1e-6), 
                 good_obj.T_dependent_property_derivative(190, order=2), rtol=1e-7)
    assert_close(derivative(lambda T:good_obj.T_dependent_property_derivative(T, 2), 190, dx=190*1e-6), 
                 good_obj.T_dependent_property_derivative(190, order=3), rtol=1e-7)
    
    
    assert_close(good_obj.T_dependent_property_derivative(50), -0.00012114625271322606, rtol=1e-11)
    assert_close(good_obj.T_dependent_property_derivative(50, 2), 7.405594884975754e-08, rtol=1e-11)
    assert_close(good_obj.T_dependent_property_derivative(50, 3), 7.578114176744999e-11, rtol=1e-11)



    good_obj = SurfaceTension(Tc=Tc, DIPPR106_parameters={'Test': {'Tmin': Tmin,
                                'Tmax': Tmax, 'Tc': Tc, 'A': A, 'B': B}},
                              extrapolation='DIPPR106_AB')
    assert_close(good_obj(591), good_obj.calculate(591, good_obj.method), rtol=1e-13)
    assert_close(good_obj(50), good_obj.calculate(50, good_obj.method), rtol=1e-13)
    
    assert_close(derivative(good_obj, 50, dx=50*1e-7), good_obj.T_dependent_property_derivative(50))
    
    assert_close(derivative(good_obj.T_dependent_property_derivative, 190, dx=190*1e-6), 
                 good_obj.T_dependent_property_derivative(190, order=2), rtol=1e-7)
    assert_close(derivative(lambda T:good_obj.T_dependent_property_derivative(T, 2), 190, dx=190*1e-6), 
                 good_obj.T_dependent_property_derivative(190, order=3), rtol=1e-7)
    
    
    assert_close(good_obj.T_dependent_property_derivative(50), -0.00012128895314509345, rtol=1e-11)
    assert_close(good_obj.T_dependent_property_derivative(50, 2), 7.154371429756683e-08, rtol=1e-11)
    assert_close(good_obj.T_dependent_property_derivative(50, 3), 8.967691377390834e-11, rtol=1e-11)

def test_SurfaceTensionMixture():
    # ['pentane', 'dichloromethane']
    T = 298.15
    P = 101325.0
    zs = [.1606, .8394]

    MWs = [72.14878, 84.93258]
    ws = zs_to_ws(zs, MWs=MWs)

    VolumeLiquids = [VolumeLiquid(CASRN="109-66-0", MW=72.14878, Tb=309.21, Tc=469.7, Pc=3370000.0, Vc=0.000311, Zc=0.26837097540904814, omega=0.251, dipole=0.0, extrapolation="constant", method=POLY_FIT, poly_fit=(144, 459.7000000000001, [1.0839519373491257e-22, -2.420837244222272e-19, 2.318236501104612e-16, -1.241609625841306e-13, 4.0636406847721776e-11, -8.315431504053525e-09, 1.038485128954003e-06, -7.224842789857136e-05, 0.0022328080060137396])),
     VolumeLiquid(CASRN="75-09-2", MW=84.93258, Tb=312.95, Tc=508.0, Pc=6350000.0, Vc=0.000177, Zc=0.26610258553203137, omega=0.2027, dipole=1.6, extrapolation="constant", method=POLY_FIT, poly_fit=(178.01, 484.5, [1.5991056738532454e-23, -3.92303910541969e-20, 4.1522438881104836e-17, -2.473595776587317e-14, 9.064684097377694e-12, -2.0911320815626796e-09, 2.9653069375266426e-07, -2.3580713574913447e-05, 0.0008567355308938564]))]

    SurfaceTensions = [SurfaceTension(CASRN="109-66-0", MW=72.14878, Tb=309.21, Tc=469.7, Pc=3370000.0, Vc=0.000311, Zc=0.26837097540904814, omega=0.251, StielPolar=0.005164116344598568, Hvap_Tb=357736.5860890071, extrapolation=None, method=VDI_PPDS),
                       SurfaceTension(CASRN="75-09-2", MW=84.93258, Tb=312.95, Tc=508.0, Pc=6350000.0, Vc=0.000177, Zc=0.26610258553203137, omega=0.2027, StielPolar=-0.027514125341022044, Hvap_Tb=333985.7240672881, extrapolation=None, method=VDI_PPDS)]

    obj = SurfaceTensionMixture(MWs=MWs, Tbs=[309.21, 312.95], Tcs=[469.7, 508.0], correct_pressure_pure=False, CASs=['109-66-0', '75-09-2'], SurfaceTensions=SurfaceTensions, VolumeLiquids=VolumeLiquids)

    sigma_wsd = obj.calculate(T, P, zs, method=WINTERFELDSCRIVENDAVIS, ws=ws)
    assert_close(sigma_wsd, 0.02388914831076298)

    # Check the default calculation method is still WINTERFELDSCRIVENDAVIS
    sigma = obj.mixture_property(T, P, zs)
    assert_close(sigma, sigma_wsd)

    # Check the other implemented methods
    sigma = obj.calculate(T, P, zs, ws, LINEAR)
    assert_close(sigma, 0.025332871945242523)

    sigma = obj.calculate(T, P, zs, ws, DIGUILIOTEJA)
    assert_close(sigma, 0.025262398831653664)

    with pytest.raises(Exception):
        obj.test_method_validity(T, P, zs, ws, 'BADMETHOD')
    with pytest.raises(Exception):
        obj.calculate(T, P, zs, ws, 'BADMETHOD')

    hash0 = hash(obj)
    obj2 = SurfaceTensionMixture.from_json(json.loads(json.dumps(obj.as_json())))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0

    obj2 = eval(str(obj))
    assert obj == obj2
    assert hash(obj) == hash0
    assert hash(obj2) == hash0




# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.'''

import numpy as np
import pytest
from thermo import *
from fluids.constants import R
from math import log, exp, sqrt, log10
from fluids.numerics import linspace, derivative, logspace, assert_close, assert_close1d, assert_close2d, assert_close3d

# TODO: add unit tests for solid phase identification and sorting

def test_vapor_score_Tpc():
    score = vapor_score_Tpc(T=300.0, Tcs=[304.2, 507.6], zs=[0.21834418746784942, 0.7816558125321506])
    assert_close(score, -163.18879226903942)

def test_vapor_score_Vpc():
    score = vapor_score_Vpc(V=0.00011316308855449715, Vcs=[9.4e-05, 0.000368], zs=[0.21834418746784942, 0.7816558125321506])
    assert_close(score, -0.000195010604079)

def test_vapor_score_Tpc_weighted():
    score = vapor_score_Tpc_weighted(T=300.0, Tcs=[304.2, 507.6], Vcs=[9.4e-05, 0.000368], zs=[0.9752234962374878, 0.024776503762512052])
    assert_close(score, -22.60037521107)

def test_vapor_score_Tpc_Vpc():
    score = vapor_score_Tpc_Vpc(T=300.0, V=0.00011316308855449715, Tcs=[304.2, 507.6], Vcs=[9.4e-05, 0.000368], zs=[0.21834418746784942, 0.7816558125321506])
    assert_close(score, -55.932094761)

def test_vapor_score_Wilson():
    # 1 component
    score = vapor_score_Wilson(T=300.0, P=1e6, zs=[1], Tcs=[304.2], Pcs=[7376460.0], omegas=[0.2252])
    assert_close(score, -5727363.494462478)


    score = vapor_score_Wilson(T=206.40935716944634, P=100.0, zs=[0.5, 0.5], Tcs=[304.2, 507.6], Pcs=[7376460.0, 3025000.0], omegas=[0.2252, 0.2975])
    assert_close(score, 1.074361930956633)

def test_vapor_score_Poling():
    assert_close(vapor_score_Poling(1.0054239121594122e-05), 1.0137457789955244)

def test_vapor_score_PIP():
    score = vapor_score_PIP(0.024809176851423774, 337.0119286073647, -4009021.959558917, 321440573.3615088, -13659.63987996052)
    assert_close(score, 0.016373735005)

def test_vapor_score_Bennett_Schmidt():
    assert_close(vapor_score_Bennett_Schmidt(7.558572848883679e-06), -7.558572848883679e-06)

def test_vapor_score_traces():
    score = vapor_score_traces(zs=[.218, .782], Tcs=[304.2, 507.6], CASs=['124-38-9', '110-54-3'])
    assert_close(score, 0.218)

    score = vapor_score_traces(zs=[.975, .025], Tcs=[304.2, 507.6], CASs=['124-38-9', '110-54-3'])
    assert_close(score, 0.975)

def test_identity_phase_states_basic():
    constants = ChemicalConstantsPackage(Tcs=[563.0, 647.14, 514.0], Vcs=[0.000274, 5.6e-05, 0.000168], Pcs=[4414000.0, 22048320.0, 6137000.0], omegas=[0.59, 0.344, 0.635], MWs=[74.1216, 18.01528, 46.06844], CASs=['71-36-3', '7732-18-5', '64-17-5'])
    properties = PropertyCorrelationsPackage(constants=constants, skip_missing=True,
                                             HeatCapacityGases=[HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [-3.787200194613107e-20, 1.7692887427654656e-16, -3.445247207129205e-13, 3.612771874320634e-10, -2.1953250181084466e-07, 7.707135849197655e-05, -0.014658388538054169, 1.5642629364740657, -7.614560475001724])),
                                             HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [5.543665000518528e-22, -2.403756749600872e-18, 4.2166477594350336e-15, -3.7965208514613565e-12, 1.823547122838406e-09, -4.3747690853614695e-07, 5.437938301211039e-05, -0.003220061088723078, 33.32731489750759])),
                                             HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [-1.162767978165682e-20, 5.4975285700787494e-17, -1.0861242757337942e-13, 1.1582703354362728e-10, -7.160627710867427e-08, 2.5392014654765875e-05, -0.004732593693568646, 0.5072291035198603, 20.037826650765965])),], )

    eos_kwargs = dict(Tcs=constants.Tcs, Pcs=constants.Pcs, omegas=constants.omegas)
    gas = CEOSGas(SRKMIX, eos_kwargs, HeatCapacityGases=properties.HeatCapacityGases)
    liq = CEOSLiquid(SRKMIX, eos_kwargs, HeatCapacityGases=properties.HeatCapacityGases)
    # flashN = FlashVLN(constants, properties, liquids=[liq, liq], gas=gas)
    T, P = 361, 1e5
    # flashN.flash(T=361, P=1e5, zs=[.25, 0.7, .05]).phase_count
    gas = gas.to(T=T, P=P, zs=[0.2384009970908655, 0.5786839935180925, 0.1829150093910419])
    liq0 = liq.to(T=T, P=P, zs=[7.619975052238032e-05, 0.9989622883894993, 0.0009615118599781474])
    liq1 = liq.to(T=T, P=P, zs=[0.6793120076703771, 0.19699746328631124, 0.12369052904331178])

    VLL_methods_here = [VL_ID_VPC, VL_ID_TPC_VPC,
                     VL_ID_POLING, VL_ID_PIP, VL_ID_BS, VL_ID_TRACES]
    LLL_methods_here = [VL_ID_TPC, VL_ID_TPC_VC_WEIGHTED, VL_ID_WILSON]

    for skip_solids in (True, False):
        for m in VLL_methods_here:
            sln = identity_phase_states(phases=[gas, liq0, liq1], constants=constants, correlations=properties,VL_method=m, skip_solids=skip_solids)
            assert sln[0] is gas
            assert sln[1][0] is liq0
            assert sln[1][1] is liq1
            assert not sln[2]

        for m in LLL_methods_here:
            sln = identity_phase_states(phases=[gas, liq0, liq1], constants=constants, correlations=properties,VL_method=m, skip_solids=skip_solids)
            assert sln[0] is None
            assert sln[1][0] is gas
            assert sln[1][1] is liq0
            assert sln[1][2] is liq1
            assert not sln[2]



        betas = [0.027939322463018015, 0.6139152961492583, 0.35814538138772367]
        phases=[gas, liq0, liq1]

    for skip_solids in (True, False):
        for m in VLL_methods_here:
            settings = BulkSettings(VL_ID=m)
            sln = identify_sort_phases(phases=phases, betas=betas, constants=constants, correlations=properties,
                                     settings=settings, skip_solids=skip_solids)
            assert sln[0] is gas
            assert sln[1][0] is liq0
            assert sln[1][1] is liq1
            assert not sln[2]
            assert_close1d(sln[3], betas)

        for m in LLL_methods_here:
            settings = BulkSettings(VL_ID=m)
            sln = identify_sort_phases(phases=phases, betas=betas, constants=constants, correlations=properties,
                                     settings=settings, skip_solids=skip_solids)
            assert sln[0] is None
            assert sln[1][2] is gas
            assert sln[1][0] is liq0
            assert sln[1][1] is liq1
            assert not sln[2]
            assert_close1d(sln[3], [0.6139152961492583, 0.35814538138772367, 0.027939322463018015])

def test_sort_phases_liquids():
    from thermo.phase_identification import VL_ID_METHODS, PROP_SORT, DENSITY_MASS, DENSITY, ISOTHERMAL_COMPRESSIBILITY, HEAT_CAPACITY
    constants = ChemicalConstantsPackage(Tcs=[563.0, 647.14, 514.0], Vcs=[0.000274, 5.6e-05, 0.000168], Pcs=[4414000.0, 22048320.0, 6137000.0], omegas=[0.59, 0.344, 0.635], MWs=[74.1216, 18.01528, 46.06844], CASs=['71-36-3', '7732-18-5', '64-17-5'])
    properties = PropertyCorrelationsPackage(constants=constants, skip_missing=True,
                                             HeatCapacityGases=[HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [-3.787200194613107e-20, 1.7692887427654656e-16, -3.445247207129205e-13, 3.612771874320634e-10, -2.1953250181084466e-07, 7.707135849197655e-05, -0.014658388538054169, 1.5642629364740657, -7.614560475001724])),
                                             HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [5.543665000518528e-22, -2.403756749600872e-18, 4.2166477594350336e-15, -3.7965208514613565e-12, 1.823547122838406e-09, -4.3747690853614695e-07, 5.437938301211039e-05, -0.003220061088723078, 33.32731489750759])),
                                             HeatCapacityGas(load_data=False, poly_fit=(50.0, 1000.0, [-1.162767978165682e-20, 5.4975285700787494e-17, -1.0861242757337942e-13, 1.1582703354362728e-10, -7.160627710867427e-08, 2.5392014654765875e-05, -0.004732593693568646, 0.5072291035198603, 20.037826650765965])),], )

    eos_kwargs = dict(Tcs=constants.Tcs, Pcs=constants.Pcs, omegas=constants.omegas)
    gas = CEOSGas(SRKMIX, eos_kwargs, HeatCapacityGases=properties.HeatCapacityGases)
    liq = CEOSLiquid(SRKMIX, eos_kwargs, HeatCapacityGases=properties.HeatCapacityGases)
    # flashN = FlashVLN(constants, properties, liquids=[liq, liq], gas=gas)
    T, P = 361, 1e5
    # flashN.flash(T=361, P=1e5, zs=[.25, 0.7, .05]).phase_count
    gas = gas.to(T=T, P=P, zs=[0.2384009970908655, 0.5786839935180925, 0.1829150093910419])
    liq0 = liq.to(T=T, P=P, zs=[7.619975052238032e-05, 0.9989622883894993, 0.0009615118599781474])
    liq1 = liq.to(T=T, P=P, zs=[0.6793120076703771, 0.19699746328631124, 0.12369052904331178])
    liq0.constants = liq1.constants = constants

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY_MASS, phase_sort_higher_first=False)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [650.9479399573546, 717.583719673216])

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY_MASS, phase_sort_higher_first=True)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [717.583719673216, 650.9479399573546])

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY, phase_sort_higher_first=True)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [717.583719673216, 650.9479399573546])

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY, phase_sort_higher_first=False)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [650.9479399573546, 717.583719673216])

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=ISOTHERMAL_COMPRESSIBILITY, phase_sort_higher_first=False)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [717.583719673216, 650.9479399573546])

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=ISOTHERMAL_COMPRESSIBILITY, phase_sort_higher_first=True)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [650.9479399573546, 717.583719673216])

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=HEAT_CAPACITY, phase_sort_higher_first=False)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [717.583719673216, 650.9479399573546])

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=HEAT_CAPACITY, phase_sort_higher_first=True)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close1d([i.rho_mass() for i in liquids], [650.9479399573546, 717.583719673216])

    # Water settings
    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY_MASS, phase_sort_higher_first=False,
                           water_sort=WATER_FIRST)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close(liquids[0].zs[1], 0.9989622883894993)

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY_MASS, phase_sort_higher_first=True,
                           water_sort=WATER_LAST)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close(liquids[1].zs[1], 0.9989622883894993)

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY_MASS, phase_sort_higher_first=True,
                           water_sort=WATER_NOT_SPECIAL)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close(liquids[0].zs[1], 0.9989622883894993)

    settings = BulkSettings(liquid_sort_method=PROP_SORT, liquid_sort_prop=DENSITY_MASS, phase_sort_higher_first=False,
                           water_sort=WATER_NOT_SPECIAL)
    liquids, _ = sort_phases(liquids=[liq0, liq1], solids=[], constants=constants, settings=settings)
    assert_close(liquids[1].zs[1], 0.9989622883894993)



# -*- coding: utf-8 -*-
'''Chemical Engineering Design Library (ChEDL). Utilities for process modeling.
Copyright (C) 2017, 2018, 2019, 2020 Caleb Bell <Caleb.Andrew.Bell@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

'''
from chemicals.elements import simple_formula_parser

rdkit_missing = 'RDKit is not installed; it is required to use this functionality'

loaded_rdkit = False
Chem, Descriptors, AllChem, rdMolDescriptors = None, None, None, None
def load_rdkit_modules():
    global loaded_rdkit, Chem, Descriptors, AllChem, rdMolDescriptors
    if loaded_rdkit:
        return
    try:
        from rdkit import Chem
        from rdkit.Chem import Descriptors
        from rdkit.Chem import AllChem
        from rdkit.Chem import rdMolDescriptors
        loaded_rdkit = True
    except:
        if not loaded_rdkit: # pragma: no cover
            raise Exception(rdkit_missing)

def smarts_fragment_priority(catalog, rdkitmol=None, smi=None):
    r'''Fragments a molecule into a set of unique groups and counts as
    specified by the `catalog`, which is a list of objects containing
    the attributes `smarts`, `group`, and `priority`.
    
    The molecule can either be an rdkit
    molecule object, or a smiles string which will be parsed by rdkit.
    Returns a dictionary of groups and their counts according to the
    priorities of the catalog provided, as well as some other information

    Parameters
    ----------
    catalog : list
        List of objects, [-]
    rdkitmol : mol, optional
        Molecule as rdkit object, [-]
    smi : str, optional
        Smiles string representing a chemical, [-]

    Returns
    -------
    counts : dict
        Dictionaty of integer counts of the found groups only, indexed by
        the `group` [-]
    group_assignments : dict[`group`: [(matched_atoms_0,), (matched_atoms_0,)]]
        A dictionary of which atoms were included in each of the groups identified, [-]
    matched_atoms : set
        A set of all of the atoms which were matched, [-]
    success : bool
        Whether or not molecule was fully fragmented, [-]
    status : str
        A string holding an explanation of why the molecule failed to be
        fragmented, if it fails; 'OK' if it suceeds.

    Notes
    -----
    Raises an exception if rdkit is not installed, or `smi` or `rdkitmol` is
    not defined.

    Examples
    --------
    '''
    if not loaded_rdkit:
        load_rdkit_modules()
    if rdkitmol is None and smi is None:
        raise Exception('Either an rdkit mol or a smiles string is required')
    if smi is not None:
        rdkitmol = Chem.MolFromSmiles(smi)
        if rdkitmol is None:
            status = 'Failed to construct mol'
            success = False
            return {}, success, status
    from itertools import combinations
    
    # Remove this
    catalog = [i for i in catalog if i.priority is not None]
    
    rdkitmol_Hs = Chem.AddHs(rdkitmol)
    # H_count = rdkitmol_Hs.GetNumAtoms() - rdkitmol.GetNumAtoms()
    atoms = simple_formula_parser(rdMolDescriptors.CalcMolFormula(rdkitmol))
    H_count = atoms.get('H', 0)

    H_counts_by_idx = {}
    for at in rdkitmol.GetAtoms():
        H_counts_by_idx[at.GetIdx()] = at.GetTotalNumHs(includeNeighbors=True)


    all_atom_idxs = set(i.GetIdx() for i in rdkitmol.GetAtoms())
    atom_count = len(all_atom_idxs)
    status = 'OK'
    success = True
    
    counts = {}
    all_matches = {}
    for obj in catalog:
        patt = obj.smart_rdkit
        if patt is None:
            smart = obj.smarts
            if isinstance(smart, (list, tuple)):
                patt = [Chem.MolFromSmarts(s) for s in smart]
            else:
                patt = Chem.MolFromSmarts(smart)
            obj.smart_rdkit = patt

        key = obj.group_id
        if isinstance(patt, (list, tuple)):
            hits = set()
            for p in patt:
                hits.update(list(rdkitmol.GetSubstructMatches(p)))
            hits = list(hits)
        else:
            hits = list(rdkitmol.GetSubstructMatches(patt))
        if hits:
            all_matches[key] = hits
            counts[key] = len(hits)

    # Higher should be lower
    priorities = [i.priority for i in catalog]
    groups = [i.group_id for i in catalog]
    group_to_obj = {o.group_id: o for o in catalog}
    catalog_by_priority =  [group_to_obj[g] for _, g in sorted(zip(priorities, groups), reverse=True)]

    all_heavies_matched_by_a_pattern = set()
    for v in all_matches.values():
        for t in v:
            all_heavies_matched_by_a_pattern.update(t)

    # excludes H
    
    ignore_matches = set()
    matched_atoms, final_group_counts, final_assignments = run_match(catalog_by_priority, all_matches, ignore_matches, all_atom_idxs, H_count)
    # Count the hydrogens

    heavy_atom_matched = atom_count == len(matched_atoms)
    hydrogens_found = 0
    for found_group in final_assignments.keys():
        for found_atoms in final_assignments[found_group]:
            if group_to_obj[found_group].hydrogen_from_smarts:
                hydrogens_found += sum(H_counts_by_idx[i] for i in found_atoms)
            else:
                hydrogens_found += group_to_obj[found_group].atoms.get('H', 0)

    #hydrogens_found = sum(group_to_obj[g].atoms.get('H', 0)*v for g, v in final_group_counts.items())
    hydrogens_matched = hydrogens_found == H_count

    if len(all_heavies_matched_by_a_pattern) != atom_count:
        status = 'Did not match all atoms present'
        success = False
        return final_group_counts, final_assignments, matched_atoms, success, status

    success = heavy_atom_matched and hydrogens_matched
    if not success:
        things_to_ignore = []
        for k in all_matches:
            for v in all_matches[k]:
                things_to_ignore.append((k, v))
                
        # if len(things_to_ignore) < 25:
            # remove_up_to = 4
        # elif len(things_to_ignore) < 25:
        #     remove_up_to = 3
        # elif len(things_to_ignore) < 18:
        #     remove_up_to = 2
        # else:
            # remove_up_to = 3
        remove_up_to = 4
        max_tries = 20000
        tries = 0

        done = False
        for remove in range(1, remove_up_to+1):
            if done:
                break
            for ignore_matches in combinations(things_to_ignore, remove):
                tries += 1
                if tries > max_tries:
                    break
                
                ignore_matches = set(ignore_matches)
                matched_atoms, final_group_counts, final_assignments = run_match(catalog_by_priority, all_matches, ignore_matches, all_atom_idxs, H_count)
                heavy_atom_matched = atom_count == len(matched_atoms)
                if not heavy_atom_matched:
                    continue

                hydrogens_found = 0
                for found_group in final_assignments.keys():
                    for found_atoms in final_assignments[found_group]:
                        if group_to_obj[found_group].hydrogen_from_smarts:
                            hydrogens_found += sum(H_counts_by_idx[i] for i in found_atoms)
                        else:
                            hydrogens_found += group_to_obj[found_group].atoms.get('H', 0)
                
                hydrogens_matched = hydrogens_found == H_count
                success = heavy_atom_matched and hydrogens_matched

                if success:
                    done = True
                    break

    if not success:
        status = 'Did not match all atoms present'

    return final_group_counts, final_assignments, matched_atoms, success, status

def run_match(catalog_by_priority, all_matches, ignore_matches, all_atom_idxs,
              H_count):
    matched_atoms = set()
    final_group_counts = {}
    final_assignments = {}
    for obj in catalog_by_priority:
        if obj.group_id in all_matches:
            for match in all_matches[obj.group_id]:
                match_set = set(match)
                if match_set.intersection(matched_atoms):
                    # At least one atom is already matched - keep looking
                    continue

                # If the group matches everything, check the group has the right number of hydrogens
                if match_set == all_atom_idxs and H_count and obj.atoms.get('H', 0) != H_count:
                    continue
                
                if (obj.group_id, match) in ignore_matches:
                    continue
                matched_atoms.update(match)
                try:
                    final_group_counts[obj.group_id]
                except:
                    final_group_counts[obj.group_id] = 0
                final_group_counts[obj.group_id] += 1
                
                try:
                    final_assignments[obj.group_id]
                except:
                    final_assignments[obj.group_id] = []
                final_assignments[obj.group_id].append(match)
    return matched_atoms, final_group_counts, final_assignments
    


def smarts_fragment(catalog, rdkitmol=None, smi=None, deduplicate=True):
    r'''Fragments a molecule into a set of unique groups and counts as
    specified by the `catalog`. The molecule can either be an rdkit
    molecule object, or a smiles string which will be parsed by rdkit.
    Returns a dictionary of groups and their counts according to the
    indexes of the catalog provided.

    Parameters
    ----------
    catalog : dict
        Dictionary indexed by keys pointing to smarts strings, [-]
    rdkitmol : mol, optional
        Molecule as rdkit object, [-]
    smi : str, optional
        Smiles string representing a chemical, [-]

    Returns
    -------
    counts : dict
        Dictionaty of integer counts of the found groups only, indexed by
        the same keys used by the catalog [-]
    success : bool
        Whether or not molecule was fully and uniquely fragmented, [-]
    status : str
        A string holding an explanation of why the molecule failed to be
        fragmented, if it fails; 'OK' if it suceeds.

    Notes
    -----
    Raises an exception if rdkit is not installed, or `smi` or `rdkitmol` is
    not defined.

    Examples
    --------
    Acetone:

    >>> smarts_fragment(catalog=J_BIGGS_JOBACK_SMARTS_id_dict, smi='CC(=O)C') # doctest:+SKIP
    ({1: 2, 24: 1}, True, 'OK')

    Sodium sulfate, (Na2O4S):

    >>> smarts_fragment(catalog=J_BIGGS_JOBACK_SMARTS_id_dict, smi='[O-]S(=O)(=O)[O-].[Na+].[Na+]') # doctest:+SKIP
    ({29: 4}, False, 'Did not match all atoms present')

    Propionic anhydride (C6H10O3):

    >>> smarts_fragment(catalog=J_BIGGS_JOBACK_SMARTS_id_dict, smi='CCC(=O)OC(=O)CC') # doctest:+SKIP
    ({1: 2, 2: 2, 28: 2}, False, 'Matched some atoms repeatedly: [4]')
    '''
    if not loaded_rdkit:
        load_rdkit_modules()
    if rdkitmol is None and smi is None:
        raise Exception('Either an rdkit mol or a smiles string is required')
    if smi is not None:
        rdkitmol = Chem.MolFromSmiles(smi)
        if rdkitmol is None:
            status = 'Failed to construct mol'
            success = False
            return {}, success, status
    from collections import Counter

    atom_count = len(rdkitmol.GetAtoms())
    status = 'OK'
    success = True

    counts = {}
    all_matches = {}
    for key, smart in catalog.items():
        if isinstance(smart, str):
            patt = Chem.MolFromSmarts(smart)
        else:
            patt = smart
        hits = list(rdkitmol.GetSubstructMatches(patt))
        if hits:
            all_matches[key] = hits
            counts[key] = len(hits)

    # Duplicate group cleanup
    matched_atoms = []
    for i in all_matches.values():
        for j in i:
            matched_atoms.extend(j)

    if deduplicate:
        dups = [i for i, c in Counter(matched_atoms).items() if c > 1]
        iteration = 0
        while (dups and iteration < 100):
            dup = dups[0]

            dup_smart_matches = []
            for group, group_match_list in all_matches.items():
                for i, group_match_i in enumerate(group_match_list):
                    if dup in group_match_i:
                        dup_smart_matches.append((group, i, group_match_i, len(group_match_i)))


            sizes = [i[3] for i in dup_smart_matches]
            max_size = max(sizes)
#            print(sizes, 'sizes', 'dup', dup, 'working_data', dup_smart_matches)
            if sizes.count(max_size) > 1:
                iteration += 1
#                print('BAD')
                # Two same size groups, continue, can't do anything
                continue
            else:
                # Remove matches that are not the largest
                for group, idx, positions, size in dup_smart_matches:
                    if size != max_size:
                        # Not handling the case of multiple duplicate matches right, indexes changing!!!
                        del all_matches[group][idx]
                        continue

            matched_atoms = []
            for i in all_matches.values():
                for j in i:
                    matched_atoms.extend(j)

            dups = [i for i, c in Counter(matched_atoms).items() if c > 1]
            iteration += 1

    matched_atoms = set()
    for i in all_matches.values():
        for j in i:
            matched_atoms.update(j)
    if len(matched_atoms) != atom_count:
        status = 'Did not match all atoms present'
        success = False

    # Check the atom aount again, this time looking for duplicate matches (only if have yet to fail)
    if success:
        matched_atoms = []
        for i in all_matches.values():
            for j in i:
                matched_atoms.extend(j)
        if len(matched_atoms) < atom_count:
            status = 'Matched %d of %d atoms only' %(len(matched_atoms), atom_count)
            success = False
        elif len(matched_atoms) > atom_count:
            status = 'Matched some atoms repeatedly: %s' %( [i for i, c in Counter(matched_atoms).items() if c > 1])
            success = False

    return counts, success, status

